import collections
from collections import namedtuple


def f():
    lower = 0
    Camel = 0
    CONSTANT = 0
    _ = 0
    MyObj1 = collections.namedtuple("MyObj1", ["a", "b"])
    MyObj2 = namedtuple("MyObj12", ["a", "b"])
