class bad:
    pass


class _bad:
    pass


class bad_class:
    pass


class Bad_Class:
    pass


class BAD_CLASS:
    pass


class Good:
    pass


class _Good:
    pass


class GoodClass:
    pass


class GOOD:
    pass
