try:
    pass
except:
    pass
except ValueError:
    pass

try:
    pass
except:
    pass
except ValueError:
    pass
finally:
    pass

try:
    pass
except:
    pass
except ValueError:
    pass
else:
    pass

try:
    pass
except:
    pass

try:
    pass
except ValueError:
    pass
except:
    pass

try:
    pass
except ValueError:
    pass

try:
    pass
finally:
    pass
